import operator

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc
import rasterio
import rasterio.features
from rasterio.transform import Affine
from ray.data.datatype import DataType
from ray.data.expressions import udf
from ray.data.extensions import TensorArray
from shapely.geometry.base import BaseGeometry

from ratiopath.openslide import OpenSlide
from ratiopath.tifffile import TiffFile
from ratiopath.tiling.types import ReadOverlaysArguments, ReadTilesArguments
from ratiopath.tiling.utils import (
    _pyarrow_group_indices,
    _read_openslide_tiles,
    _read_tifffile_tiles,
)


def _scale_overlay(
    slide: OpenSlide | TiffFile,
    tile_x: pa.IntegerArray,
    tile_y: pa.IntegerArray,
    tile_extent_x: pa.IntegerArray,
    tile_extent_y: pa.IntegerArray,
    mpp_x: pa.FloatArray,
    mpp_y: pa.FloatArray,
) -> ReadTilesArguments:
    """Scale overlay tile arguments based on slide and overlay resolutions.

    Args:
        slide: The opened whole-slide image.
        tile_x: X coordinates of the underlying tiles.
        tile_y: Y coordinates of the underlying tiles.
        tile_extent_x: Widths of the underlying tiles.
        tile_extent_y: Heights of the underlying tiles.
        mpp_x: Physical resolution (µm/px) of the underlying slide in X direction.
        mpp_y: Physical resolution (µm/px) of the underlying slide in Y direction.

    Returns:
        A dictionary containing the scaled tile arguments.
    """
    mpp = pd.Series(zip(mpp_x.to_numpy(), mpp_y.to_numpy(), strict=True))  # type: ignore [call-overload]
    level = mpp.apply(slide.closest_level)

    # Determine the overlay resolution
    overlay_mpp = level.apply(slide.slide_resolution)

    x_scaling = pa.array(
        mpp.apply(operator.itemgetter(0)) / overlay_mpp.apply(operator.itemgetter(0)),
        pa.float32(),
    )
    y_scaling = pa.array(
        mpp.apply(operator.itemgetter(1)) / overlay_mpp.apply(operator.itemgetter(1)),
        pa.float32(),
    )

    def scale(x: pa.IntegerArray, scaling: pa.FloatArray) -> pa.IntegerArray:
        return pc.max_element_wise(
            pa.scalar(0),
            pc.round(pc.multiply(x, scaling)).cast(pa.int32()),
        )  # type: ignore [return-value]

    return {
        "tile_x": scale(tile_x, x_scaling),
        "tile_y": scale(tile_y, y_scaling),
        "tile_extent_x": scale(tile_extent_x, x_scaling),
        "tile_extent_y": scale(tile_extent_y, y_scaling),
        "level": pa.array(level, pa.int8()),
    }


def _read_openslide_overlay(
    path: str, kwargs: ReadOverlaysArguments
) -> tuple[np.ndarray, ReadTilesArguments]:
    """Read batch of overlays from a whole-slide image using OpenSlide."""
    with OpenSlide(path) as slide:
        new_kwargs = _scale_overlay(slide, **kwargs)
        # Use a background of 0 (black) for overlays
        return _read_openslide_tiles(slide, **new_kwargs, background=0), new_kwargs


def _read_tifffile_overlay(
    path: str, kwargs: ReadOverlaysArguments
) -> tuple[np.ndarray, ReadTilesArguments]:
    """Read batch of overlays from an OME-TIFF file using tifffile."""
    with TiffFile(path) as slide:
        new_kwargs = _scale_overlay(slide, **kwargs)
        # Use a background of 0 (black) for overlays
        return _read_tifffile_tiles(slide, **new_kwargs, background=0), new_kwargs


def _tile_overlay(
    roi: BaseGeometry,
    overlay_path: pa.StringArray,
    tile_x: pa.IntegerArray,
    tile_y: pa.IntegerArray,
    mpp_x: pa.FloatArray,
    mpp_y: pa.FloatArray,
) -> np.ma.MaskedArray:
    """Read overlay tiles for a batch of tiles.

    For each overlay path the corresponding whole-slide image is opened (OpenSlide or OME-TIFF).
    The overlay is accessed at the slide level closest to each tile's mpp and the tile
    coordinates/extents are scaled to that level before reading.

    The region of interest (ROI) geometry is treated in the same image space (resolution) as the underlying slide tiles.
    The region can be an arbitrary polygon. However, a bounding box of the region is used for reading overlay tiles
    and then masked to respect the region defined by provided overlay.

    Args:
        roi: The region of interest geometry.
        overlay_path: A pyarrow array of whole-slide image paths for the overlays.
        tile_x: A pyarrow array of tile x-coordinates.
        tile_y: A pyarrow array of tile y-coordinates.
        mpp_x: A pyarrow array of physical resolutions (µm/px) of the underlying slide in X direction.
        mpp_y: A pyarrow array of physical resolutions (µm/px) of the underlying slide in Y direction.

    Returns:
        A pyarrow array of masked numpy arrays containing the read overlay tiles.
    """
    # Pre-calculate ROI constants
    roi_min_x, roi_min_y, roi_max_x, roi_max_y = map(round, roi.bounds)
    roi_width = roi_max_x - roi_min_x
    roi_height = roi_max_y - roi_min_y

    tile_x = pc.add(tile_x, pa.scalar(roi_min_x))
    tile_y = pc.add(tile_y, pa.scalar(roi_min_y))

    # Pre-allocate output array (Object array to hold MaskedArrays)
    masked_tiles = np.empty(len(tile_x), dtype=object)

    def mask_tile(
        overlay: np.ndarray, x: int, y: int, extent_x: int, extent_y: int
    ) -> np.ma.MaskedArray:
        mask = rasterio.features.geometry_mask(
            geometries=[roi],
            out_shape=overlay.shape[:2],
            # Scale and translate to tile coordinates
            transform=Affine.translation(x, y)
            * Affine.scale(extent_x / roi_width, extent_y / roi_height),
        )

        return np.ma.masked_array(overlay, mask=np.dstack([mask] * overlay.shape[2]))

    for path, group in _pyarrow_group_indices(overlay_path).items():
        assert isinstance(path, str)

        batch_x = pc.take(tile_x, group)
        batch_y = pc.take(tile_y, group)

        overlay_kwargs = ReadOverlaysArguments(
            tile_x=batch_x,
            tile_y=batch_y,
            tile_extent_x=pa.repeat(roi_width, len(group)),
            tile_extent_y=pa.repeat(roi_height, len(group)),
            mpp_x=pc.take(mpp_x, group),
            mpp_y=pc.take(mpp_y, group),
        )

        # Check if it's an OME-TIFF file
        if path.lower().endswith((".ome.tiff", ".ome.tif")):
            tiles, tile_kwargs = _read_tifffile_overlay(path, overlay_kwargs)
        else:
            tiles, tile_kwargs = _read_openslide_overlay(path, overlay_kwargs)

        batch_x: pa.IntegerArray = pc.min_element_wise(pa.scalar(roi_min_x), batch_x)  # type: ignore [no-redef]
        batch_y: pa.IntegerArray = pc.min_element_wise(pa.scalar(roi_min_y), batch_y)  # type: ignore [no-redef]

        masked_tiles[group] = [
            mask_tile(
                tile,
                batch_x[i].as_py(),
                batch_y[i].as_py(),
                tile_kwargs["tile_extent_x"][i].as_py(),
                tile_kwargs["tile_extent_y"][i].as_py(),
            )
            for i, tile in enumerate(tiles)
        ]

    return masked_tiles  # type: ignore [return-value]


@udf(return_dtype=DataType(np.ndarray))
def tile_overlay(
    roi: BaseGeometry,
    overlay_path: pa.StringArray,
    tile_x: pa.IntegerArray,
    tile_y: pa.IntegerArray,
    mpp_x: pa.FloatArray,
    mpp_y: pa.FloatArray,
) -> pa.Array:
    """Read overlay tiles for a batch of tiles.

    For each overlay path the corresponding whole-slide image is opened (OpenSlide or OME-TIFF).
    The overlay is accessed at the slide level closest to each tile's mpp and the tile
    coordinates/extents are scaled to that level before reading.

    The region of interest (ROI) geometry is treated in the same image space (resolution) as the underlying slide tiles.
    The region can be an arbitrary polygon. However, a bounding box of the region is used for reading overlay tiles
    and then masked to respect the region defined by provided overlay.

    Unfortunately, at the moment we cannot use masked arrays directly in Ray Dataset. So instead of a numpy masked array,
    we provide the data and the mask as 2 separate arrays. The implementation is handled via TensorArray.

    Args:
        roi: The region of interest geometry.
        overlay_path: A pyarrow array of whole-slide image paths for the overlays.
        tile_x: A pyarrow array of tile x-coordinates.
        tile_y: A pyarrow array of tile y-coordinates.
        mpp_x: A pyarrow array of physical resolutions (µm/px) of the underlying slide in X direction.
        mpp_y: A pyarrow array of physical resolutions (µm/px) of the underlying slide in Y direction.

    Returns:
        A pyarrow array of masked numpy arrays containing the read overlay tiles.
            - The first element is the tile data.
            - The second element is the mask (True for pixels outside the ROI).
    """
    overlays = _tile_overlay(roi, overlay_path, tile_x, tile_y, mpp_x, mpp_y)

    return pa.array(TensorArray([[overlay.data, overlay.mask] for overlay in overlays]))


@udf(return_dtype=DataType(dict))
def tile_overlay_overlap(
    roi: BaseGeometry,
    overlay_path: pa.StringArray,
    tile_x: pa.IntegerArray,
    tile_y: pa.IntegerArray,
    mpp_x: pa.FloatArray,
    mpp_y: pa.FloatArray,
) -> pa.MapArray:
    """Calculate the overlap of each overlay tile with the region of interest (ROI).

    For each overlay path the corresponding whole-slide image is opened (OpenSlide or OME-TIFF).
    The overlay is accessed at the slide level closest to each tile's mpp and the tile
    coordinates/extents are scaled to that level before reading.

    The region of interest (ROI) geometry is treated in the same image space (resolution) as the underlying slide tiles.
    The region can be an arbitrary polygon.

    The Pyarrow array that is used inside ray dataset stores data in array like dictionary.
    This results in all rows having same set of keys and missing keys are filled with Nones.
    Furthermore Pyarrow only support string keys in dictionaries. Therefore the keys in the
    resulting dictionary are string representations of the overlay values.

    Args:
        roi: The region of interest geometry.
        overlay_path: A pyarrow array of whole-slide image paths for the overlays.
        tile_x: A pyarrow array of tile x-coordinates.
        tile_y: A pyarrow array of tile y-coordinates.
        mpp_x: A pyarrow array of physical resolutions (µm/px) of the underlying slide in X direction.
        mpp_y: A pyarrow array of physical resolutions (µm/px) of the underlying slide in Y direction.

    Returns:
        A pyarrow array of dictionaries mapping overlay values to their overlap fraction with the ROI.
    """
    # The overlay is a masked array where the mask is True for pixels outside the ROI.
    overlay = _tile_overlay(roi, overlay_path, tile_x, tile_y, mpp_x, mpp_y)

    def overlap_fraction(overlay: np.ma.MaskedArray) -> dict[str, float]:
        """Calculate the overlap fraction of each unique value in the overlay."""
        return {
            # Pyarrow requires string keys in dictionaries
            str(value.item()): count.item() / overlay.count()
            for value, count in zip(
                *np.unique(overlay.compressed(), return_counts=True), strict=True
            )
        }

    overlap_vectorized = np.vectorize(overlap_fraction, otypes=[object])

    return pa.array(overlap_vectorized(overlay))



from ratiopath.masks.vips_filters.color_filters import (
    VipsGrayScaleFilter,
    VipsSaturationFilter,
)
from ratiopath.masks.vips_filters.vips_filter import VipsCompose, VipsFilter
from ratiopath.masks.vips_filters.vips_morphology import VipsClosing, VipsOpening
from ratiopath.masks.vips_filters.vips_multi_otsu import VipsMultiOtsu, VipsOtsu


__all__ = [
    "VipsClosing",
    "VipsCompose",
    "VipsFilter",
    "VipsGrayScaleFilter",
    "VipsMultiOtsu",
    "VipsOpening",
    "VipsOtsu",
    "VipsSaturationFilter",
]



import pytest

from ratiopath.tiling import grid_tiles


@pytest.mark.parametrize(
    "slide_extent, tile_extent, stride, last, expected",
    [
        (
            (5, 5),
            (2, 2),
            (2, 2),
            "drop",
            [(0, 0), (2, 0), (0, 2), (2, 2)],
        ),
        (
            (5, 5),
            (2, 2),
            (2, 2),
            "keep",
            [(0, 0), (2, 0), (4, 0), (0, 2), (2, 2), (4, 2), (0, 4), (2, 4), (4, 4)],
        ),
        (
            (5, 5),
            (2, 2),
            (2, 2),
            "shift",
            [(0, 0), (2, 0), (3, 0), (0, 2), (2, 2), (3, 2), (0, 3), (2, 3), (3, 3)],
        ),
    ],
)
def test_grid_tiler(slide_extent, tile_extent, stride, last, expected):
    tiles = [
        tuple(tile) for tile in grid_tiles(slide_extent, tile_extent, stride, last)
    ]
    assert tiles == expected



import numpy as np


HEMATOXYLIN = np.array([0.65, 0.70, 0.29])  # From Ruifrok & Johnston's original paper
# EOSIN = np.array([0.07, 0.99, 0.11]) # From Ruifrok & Johnston's original paper
EOSIN = np.array([0.2159, 0.8012, 0.5581])  # From http://amida13.isi.uu.nl/?q=node/69
DAB = np.array([0.27, 0.57, 0.78])  # From Ruifrok & Johnston's original paper


def make_residual_stain(stain1: np.ndarray, stain2: np.ndarray) -> np.ndarray:
    """Create a residual stain vector from two stain vectors.

    Parameters:
        stain1: A numpy array representing the first stain vector.
        stain2: A numpy array representing the second stain vector.

    Returns:
        A numpy array representing the residual stain vector.
    """
    res = np.linalg.cross(stain1, stain2)
    return res / np.linalg.norm(res)


HE = np.array(
    [HEMATOXYLIN, EOSIN, make_residual_stain(HEMATOXYLIN, EOSIN)], dtype=np.float32
)
HDAB = np.array(
    [HEMATOXYLIN, DAB, make_residual_stain(HEMATOXYLIN, DAB)], dtype=np.float32
)


def discard_pixels(
    od: np.ndarray,
    min_stain: float,
    max_stain: float,
    gray_threshold: float = np.cos(0.15),
) -> np.ndarray:
    """Discard pixels based on optical density thresholds.

    Parameters:
        od: A numpy array of optical densities for red, green, and blue channels.
        min_stain: Minimum optical density threshold.
        max_stain: Maximum optical density threshold.
        gray_threshold: Threshold for excluding very gray pixels (default is cos(0.15)).

    Returns:
        A numpy array containing the filtered optical densities for red, green, and blue channels.
    """
    keep_count = 0
    max_stain_squared = max_stain * max_stain
    sqrt3 = 1 / np.sqrt(3)

    for i in range(len(od)):
        r, g, b = od[i]
        mag_squared = r * r + g * g + b * b
        if (
            mag_squared > max_stain_squared
            or r < min_stain
            or g < min_stain
            or b < min_stain
            or mag_squared <= 0
        ):
            continue

        # Exclude very gray pixels
        if (r * sqrt3 + g * sqrt3 + b * sqrt3) / np.sqrt(mag_squared) >= gray_threshold:
            continue

        od[keep_count] = np.array([r, g, b])
        keep_count += 1

    return od[:keep_count]


def estimate_stain_vectors(
    image: np.ndarray,
    default_stain_vectors: np.ndarray,
    i0: int = 256,
    min_stain: float = 0.05,
    max_stain: float = 1.0,
    alpha: float = 0.01,
) -> np.ndarray:
    """Estimate stain vectors from an image using optical density transformation.

    Parameters:
        image: A numpy array representing the input image.
        default_stain_vectors: A numpy array of default unit stain vectors.
        i0: The intensity value for normalization.
        min_stain: Minimum optical density threshold for discarding pixels.
        max_stain: Maximum optical density threshold for discarding pixels.
        alpha: The percentage of pixels to use for estimating the stain vectors
            (default is 0.01, which corresponds to 1%).

    Returns:
        A numpy array of estimated stain vectors.

    References:
        Paper: A method for normalizing histology slides for quantitative analysis,
            M Macenko, M Niethammer, JS Marron, D Borland, JT Woosley, G Xiaojun,
            C Schmitt, NE Thomas, IEEE ISBI, 2009. dx.doi.org/10.1109/ISBI.2009.5193250
    """
    od = -np.log10(np.maximum(image.reshape(-1, 3), 1) / i0)

    od_hat = discard_pixels(od, min_stain, max_stain)

    cov = np.cov(od_hat.T)

    _, eigvecs = np.linalg.eigh(cov)

    t_hat = od_hat.dot(eigvecs[:, 1:3])

    phi = np.arctan2(t_hat[:, 1], t_hat[:, 0])

    min_phi = np.percentile(phi, alpha)
    max_phi = np.percentile(phi, 100 - alpha)

    direction_min = np.array([np.cos(min_phi), np.sin(min_phi)], dtype=np.float32)
    stain1 = eigvecs[:, 1:3].dot(direction_min).reshape(-1)
    stain1 /= np.linalg.norm(stain1)

    direction_max = np.array([np.cos(max_phi), np.sin(max_phi)], dtype=np.float32)
    stain2 = eigvecs[:, 1:3].dot(direction_max).reshape(-1)
    stain2 /= np.linalg.norm(stain2)

    cos_angle11 = np.dot(stain1, default_stain_vectors[0])
    cos_angle12 = np.dot(stain1, default_stain_vectors[1])
    cos_angle21 = np.dot(stain2, default_stain_vectors[0])
    cos_angle22 = np.dot(stain2, default_stain_vectors[1])

    if max(cos_angle12, cos_angle21) > max(cos_angle11, cos_angle22):
        stain1, stain2 = stain2, stain1

    return np.array([stain1, stain2, make_residual_stain(stain1, stain2)])



from typing import Any, Literal

import pyarrow.fs

import ray
from ratiopath.ray.datasource.slide_metadatasource import (
    SlideFileMetadataProvider,
    SlideMetaDatasource,
)
from ray.data import Dataset
from ray.data.datasource import BaseFileMetadataProvider, PathPartitionFilter
from ray.data.datasource.file_based_datasource import (
    FileShuffleConfig,
    _validate_shuffle_arg,
)
from ray.data.datasource.partitioning import Partitioning


FILE_EXTENSIONS = [
    # OpenSlide formats
    "svs",
    "tif",
    "dcm",
    "ndpi",
    "vms",
    "vmu",
    "scn",
    "mrxs",
    "tiff",
    "svslide",
    "bif",
    "czi",
    # OME-TIFF formats
    "ome.tiff",
    "ome.tif",
]


def read_slides(
    paths: str | list[str],
    *,
    tile_extent: int | tuple[int, int],
    stride: int | tuple[int, int],
    mpp: float | None = None,
    level: int | None = None,
    filesystem: pyarrow.fs.FileSystem | None = None,
    ray_remote_args: dict[str, Any] | None = None,
    meta_provider: BaseFileMetadataProvider | None = None,
    partition_filter: PathPartitionFilter | None = None,
    partitioning: Partitioning | None = None,
    shuffle: Literal["files"] | FileShuffleConfig | None = None,
    ignore_missing_paths: bool = False,
    file_extensions: list[str] | None = FILE_EXTENSIONS,
    concurrency: int | None = None,
    override_num_blocks: int | None = None,
) -> Dataset:
    """Creates a :class:`~ray.data.Dataset` from whole slide image files.

    This function reads metadata from whole slide image (WSI) files and creates a
    Ray Dataset where each row corresponds to a single slide. The dataset contains
    metadata required for subsequent tiled processing, such as slide dimensions,
    resolution (MPP), and tiling parameters.

    It automatically selects the best slide level based on the specified `mpp`
    (microns per pixel) or uses the given `level`.

    Examples:
        Read a single slide and create a metadata dataset.

        >>> import ray
        >>> from ratiopath.ray import read_slide
        >>> ds = read_slide(  # doctest: +SKIP
        ...     "path/to/slide.svs",
        ...     tile_extent=256,
        ...     stride=256,
        ...     mpp=0.5,
        ... )
        >>> ds.schema()  # doctest: +SKIP
        Column         Type
        ------         ----
        path           string
        extent_x       int64
        extent_y       int64
        tile_extent_x  int64
        tile_extent_y  int64
        stride_x       int64
        stride_y       int64
        mpp_x          double
        mpp_y          double
        level          int64
        downsample     double

    Args:
        paths: A single file path or a list of file paths to whole slide images.
        tile_extent: The size of the tiles to be generated, as `(width, height)`.
            If a single integer is provided, it's used for both dimensions.
        stride: The step size between consecutive tiles, as `(x_stride, y_stride)`.
            If a single integer is provided, it's used for both dimensions.
        mpp: The desired microns per pixel. The datasource will select the slide
            level with the closest MPP. Exactly one of `mpp` or `level` must be
            provided.
        level: The desired slide level to use. Exactly one of `mpp` or `level`
            must be provided.
        filesystem: The PyArrow filesystem implementation to read from. If not
            provided, it will be inferred from the file paths.
        ray_remote_args: kwargs passed to :func:`ray.remote` in the read tasks.
        meta_provider: Custom metadata providers may be able to resolve file metadata
            more quickly and/or accurately. In most cases you do not need to set this
            parameter.
        partition_filter: A filter to read only selected partitions of a dataset.
        partitioning: A :class:`~ray.data.datasource.partitioning.Partitioning` object
            that describes how paths are organized.
        shuffle: If set to "files", randomly shuffles the input file order.
        ignore_missing_paths: If `True`, ignores any file paths that don't exist.
        file_extensions: A list of file extensions to filter files by. If `None`,
            it uses the default list of supported slide formats.
        concurrency: The maximum number of Ray tasks to run concurrently.
        override_num_blocks: Override the number of output blocks from all read tasks.

    Returns:
        A :class:`~ray.data.Dataset` where each row contains the metadata for one
        slide, ready for tiling operations.
    """
    _validate_shuffle_arg(shuffle)

    if meta_provider is None:
        meta_provider = SlideFileMetadataProvider()

    datasource = SlideMetaDatasource(
        paths,
        tile_extent=tile_extent,
        stride=stride,
        mpp=mpp,
        level=level,
        filesystem=filesystem,
        partition_filter=partition_filter,
        partitioning=partitioning,
        ignore_missing_paths=ignore_missing_paths,
        shuffle=shuffle,
        file_extensions=file_extensions,
    )
    return ray.data.read_datasource(
        datasource,
        ray_remote_args=ray_remote_args,
        concurrency=concurrency,
        override_num_blocks=override_num_blocks,
    )



from abc import abstractmethod

import pyvips
from numpy.typing import NDArray
from skimage.filters import threshold_multiotsu

from ratiopath.masks.vips_filters.typed import Res
from ratiopath.masks.vips_filters.vips_filter import VipsFilter


class VipsMultiOtsu(VipsFilter):
    @abstractmethod
    def apply_threshold(
        self, image: pyvips.Image, thresholds: list[float]
    ) -> pyvips.Image: ...

    @abstractmethod
    def otsu_threshold(self, hist: NDArray) -> list[float]: ...

    def __call__(self, image: pyvips.Image, mpp: Res) -> tuple[pyvips.Image, Res]:
        thresholds = self.otsu_threshold(image.hist_find().numpy()[0])
        return self.apply_threshold(image, thresholds), mpp


class VipsOtsu(VipsMultiOtsu):
    """Applies Otsu thresholding to an image.

    The Otsu thresholding method is used to separate the image into two classes: background and foreground.
    We use multi-Otsu to separate the image into 4 classes as there are usually up to 4 dominant colors in the image.
    Then we take the last class as the background and the rest as the foreground.
    """

    def apply_threshold(
        self, image: pyvips.Image, thresholds: list[float]
    ) -> pyvips.Image:
        return image < thresholds[2]

    def otsu_threshold(self, hist: NDArray) -> list[float]:
        return threshold_multiotsu(hist=hist, classes=4)



import pyvips

from ratiopath.masks.vips_filters.typed import Res
from ratiopath.masks.vips_filters.vips_filter import VipsFilter


class VipsGrayScaleFilter(VipsFilter):
    def __call__(self, image: pyvips.Image, mpp: Res) -> tuple[pyvips.Image, Res]:
        return image.colourspace(pyvips.enums.Interpretation.B_W).bandsplit()[0], mpp


class VipsSaturationFilter(VipsFilter):
    def __call__(self, image: pyvips.Image, mpp: Res) -> tuple[pyvips.Image, Res]:
        # Extract saturation channel
        return image.sRGB2HSV().bandsplit()[1], mpp



"""Tests for annotation parsers."""

import io
import json

import pytest

from ratiopath.parsers import ASAPParser, Darwin7JSONParser, GeoJSONParser


class TestASAPParser:
    """Test the ASAP parser."""

    @pytest.fixture
    def asap_xml_content(self):
        """Sample ASAP XML content."""
        return """<?xml version="1.0" encoding="UTF-8"?>
        <ASAP_Annotations>
            <Annotations>
                <Annotation Name="Annotation 1" Type="Polygon" PartOfGroup="None" Color="#FF0000">
                    <Coordinates>
                        <Coordinate Order="0" X="100.0" Y="200.0" />
                        <Coordinate Order="1" X="150.0" Y="200.0" />
                        <Coordinate Order="2" X="125.0" Y="250.0" />
                    </Coordinates>
                </Annotation>
                <Annotation Name="Annotation 2" Type="Point" PartOfGroup="None" Color="#00FF00">
                    <Coordinates>
                        <Coordinate Order="0" X="300.0" Y="400.0" />
                    </Coordinates>
                </Annotation>
            </Annotations>
        </ASAP_Annotations>"""

    def test_get_polygons(self, asap_xml_content):
        """Test parsing polygons from ASAP XML."""
        f = io.StringIO(asap_xml_content)

        parser = ASAPParser(f)
        polygons = list(parser.get_polygons())

        assert len(polygons) == 1
        # Check that we have a polygon-like object
        polygon = polygons[0]
        assert hasattr(polygon, "exterior")

    def test_get_points(self, asap_xml_content):
        """Test parsing points from ASAP XML."""
        f = io.StringIO(asap_xml_content)

        parser = ASAPParser(f)
        points = list(parser.get_points())

        assert len(points) == 1
        # Check that we have a point-like object
        point = points[0]
        assert hasattr(point, "x") and hasattr(point, "y")

    def test_get_polygons_with_filters(self, asap_xml_content):
        """Test parsing polygons with filters."""
        f = io.StringIO(asap_xml_content)
        parser = ASAPParser(f)
        polygons = list(parser.get_polygons(name="Annotation 1"))

        assert len(polygons) == 1
        assert polygons[0].exterior.coords[0] == (100.0, 200.0)

        polygons = list(parser.get_polygons(name="Nonexistent"))
        assert len(polygons) == 0


class TestGeoJSONParser:
    """Test the GeoJSON parser."""

    @pytest.fixture
    def geojson_content(self):
        """Sample GeoJSON content."""
        return {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "Polygon",
                        "coordinates": [
                            [
                                [100.0, 200.0],
                                [150.0, 200.0],
                                [125.0, 250.0],
                                [100.0, 200.0],
                            ]
                        ],
                    },
                    "properties": {"nested": {"property": "value"}},
                },
                {
                    "type": "Feature",
                    "geometry": {"type": "Point", "coordinates": [300.0, 400.0]},
                    "properties": {"name": "point1"},
                },
            ],
        }

    def test_get_polygons(self, geojson_content):
        """Test parsing polygons from GeoJSON."""
        f = io.StringIO(json.dumps(geojson_content))

        parser = GeoJSONParser(f)
        polygons = list(parser.get_polygons())

        assert len(polygons) == 1
        # Check that we have a polygon-like object
        polygon = polygons[0]
        assert hasattr(polygon, "exterior")

    def test_get_points(self, geojson_content):
        """Test parsing points from GeoJSON."""
        f = io.StringIO(json.dumps(geojson_content))

        parser = GeoJSONParser(f)
        points = list(parser.get_points())

        assert len(points) == 1
        # Check that we have a point-like object
        point = points[0]
        assert hasattr(point, "x") and hasattr(point, "y")

    def test_get_polygons_with_filters(self, geojson_content):
        """Test parsing polygons with filters."""
        f = io.StringIO(json.dumps(geojson_content))

        parser = GeoJSONParser(f)
        polygons = list(parser.get_polygons(nested_property="value"))

        assert len(polygons) == 1
        assert polygons[0].exterior.coords[0] == (100.0, 200.0)

        polygons = list(parser.get_polygons(name="nested"))
        assert len(polygons) == 0

        polygons = list(parser.get_polygons(name="nonexistent"))
        assert len(polygons) == 0


class TestDarwin7JSONParser:
    """Test the Darwin JSON parser."""

    @pytest.fixture
    def darwin_json_content(self):
        """Sample Darwin JSON V2 content."""
        return {
            "version": "2.0",
            "item": {"name": "bst-001.tiff"},
            "annotations": [
                {
                    "id": "1",
                    "name": "EPITHELIUM",
                    "polygon": {
                        "paths": [
                            [
                                {"x": 40702.0455, "y": 126600.7015},
                                {"x": 40651.8178, "y": 126669.5851},
                                {"x": 40650.3827, "y": 126669.5851},
                            ]
                        ]
                    },
                    "properties": [],
                },
                {
                    "id": "2",
                    "name": "POINT-ANNOTATION",
                    "point": {"x": 300.0, "y": 400.0},
                    "properties": [],
                },
            ],
        }

    def test_get_polygons(self, darwin_json_content):
        f = io.StringIO(json.dumps(darwin_json_content))
        parser = Darwin7JSONParser(f)
        polygons = list(parser.get_polygons())

        assert len(polygons) == 1
        assert hasattr(polygons[0], "exterior")
        assert polygons[0].exterior.coords[0] == (40702.0455, 126600.7015)

    def test_get_points(self, darwin_json_content):
        f = io.StringIO(json.dumps(darwin_json_content))
        parser = Darwin7JSONParser(f)
        points = list(parser.get_points())

        assert len(points) == 1
        assert hasattr(points[0], "x") and hasattr(points[0], "y")
        assert points[0].x == 300.0
        assert points[0].y == 400.0

    def test_get_polygons_with_filters(self, darwin_json_content):
        f = io.StringIO(json.dumps(darwin_json_content))
        parser = Darwin7JSONParser(f)

        polygons = list(parser.get_polygons(name="EPITHELIUM"))
        assert len(polygons) == 1

        polygons = list(parser.get_polygons(name="Nonexistent"))
        assert len(polygons) == 0



import numbers
from collections.abc import Iterator
from itertools import chain
from typing import Any, TypeAlias

import numpy as np
import pandas as pd
from scipy.sparse import spmatrix
from sklearn.model_selection import (
    BaseShuffleSplit,
    GroupShuffleSplit,
    ShuffleSplit,
    StratifiedGroupKFold,
    StratifiedShuffleSplit,
)
from sklearn.model_selection._split import GroupsConsumerMixin, _validate_shuffle_split
from sklearn.utils._array_api import get_namespace_and_device, move_to
from sklearn.utils._indexing import _safe_indexing
from sklearn.utils._param_validation import Interval, RealNotInt, validate_params
from sklearn.utils.validation import _num_samples, check_random_state, indexable


ArrayLike: TypeAlias = np.typing.ArrayLike
MatrixLike: TypeAlias = np.ndarray | pd.DataFrame | spmatrix
Int: TypeAlias = int | np.int8 | np.int16 | np.int32 | np.int64
Float: TypeAlias = float | np.float16 | np.float32 | np.float64


class StratifiedGroupShuffleSplit(GroupsConsumerMixin, BaseShuffleSplit):
    """Stratified shuffle split with non-overlapping groups.

    Provides train/test indices to split data such that both stratification
    (preserving class distribution) and grouping (non-overlapping groups between
    splits) are maintained.

    This splitter combines the functionality of StratifiedShuffleSplit and
    GroupShuffleSplit. It attempts to create folds which preserve the percentage
    of samples from each class while ensuring that samples from the same group
    do not appear in both train and test sets.

    Read more in the :ref:`User Guide <cross_validation>`.

    Parameters:
        n_splits: Number of re-shuffling & splitting iterations.
        test_size: If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split. If int, represents the absolute
            number of test samples. If None, the value is set to the complement of the
            train size.
        train_size: If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the train split. If int, represents the
            absolute number of train samples. If None, the value is automatically set to
            the complement of the test size.
        random_state: Controls the randomness of the training and testing indices. Pass
            an int for reproducible output across multiple function calls.
            See :term:`Glossary <random_state>`.

    Examples:
        >>> import numpy as np
        >>> from ratiopath.model_selection import StratifiedGroupShuffleSplit
        >>> X = np.array([[1, 2], [3, 4], [5, 6], [7, 8], [9, 10], [11, 12]])
        >>> y = np.array([0, 0, 1, 1, 0, 1])
        >>> groups = np.array([1, 1, 2, 2, 3, 3])
        >>> sgss = StratifiedGroupShuffleSplit(n_splits=2, random_state=42)
        >>> for train_index, test_index in sgss.split(X, y, groups):
        ...     print(f"Train: {train_index}, Test: {test_index}")
        Train: [0 1 2 3], Test: [4 5]
        Train: [2 3 4 5], Test: [0 1]

    Notes:
        The implementation finds the best stratification split by trying multiple splits
        and selecting the one that minimizes the difference between the class
        distributions in the original data and the test split.
    """

    def __init__(
        self,
        n_splits: Int = 5,
        *,
        test_size: None | Float = None,
        train_size: None | Float = None,
        random_state: np.random.RandomState | None | Int = None,
    ) -> None:
        super().__init__(
            n_splits=n_splits,
            test_size=test_size,
            train_size=train_size,
            random_state=random_state,
        )
        self._default_test_size = 0.2

    @staticmethod
    def _get_distribution(labels: ArrayLike) -> np.ndarray:
        _, counts = np.unique(labels, return_counts=True)
        return counts / counts.sum()

    def split(
        self,
        X: list[str] | MatrixLike,  # noqa: N803
        y: ArrayLike | None = None,
        groups: Any = None,
    ) -> Iterator[Any]:
        """Generate indices to split data into training and test set.

        Parameters:
            X: Training data, where ``n_samples`` is the number of samples and
                ``n_features`` is the number of features.
            y: The target variable for supervised learning problems. Stratification is
                done based on the y labels.
            groups: Group labels for the samples used while splitting the dataset into
                train and test set. Must be provided.

        Yields:
            train: The training set indices for that split.
            test: The testing set indices for that split.
        """
        n_samples = _num_samples(X)
        n_train, n_test = _validate_shuffle_split(
            n_samples, self.test_size, self.train_size, self._default_test_size
        )

        flipped = False
        if n_test > n_train:
            # Approximation using folds is terrible when the test set is larger than the train set
            n_test, n_train = n_train, n_test
            flipped = True

        n_splits = round(n_samples / n_test)
        rng = check_random_state(self.random_state)
        y = np.asarray(y)
        data_distribution = self._get_distribution(y)

        for _ in range(self.n_splits):
            min_diff: Float | None = None
            train_index: np.ndarray | None = None
            test_index: np.ndarray | None = None
            cv = StratifiedGroupKFold(n_splits=n_splits, shuffle=True, random_state=rng)

            for curr_train_index, curr_test_index in cv.split(X=X, y=y, groups=groups):
                test_distribution = self._get_distribution(y[curr_test_index])

                if len(test_distribution) == len(data_distribution):
                    diff = np.abs(test_distribution - data_distribution).sum()
                else:
                    diff = float("inf")

                if min_diff is None or diff < min_diff:
                    min_diff = diff
                    train_index = curr_train_index
                    test_index = curr_test_index

            if flipped:
                train_index, test_index = test_index, train_index
            yield train_index, test_index


# https://github.com/scikit-learn/scikit-learn/blob/d3898d9d5/sklearn/model_selection/_split.py#L2757
@validate_params(
    {
        "test_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(numbers.Integral, 1, None, closed="left"),
            None,
        ],
        "train_size": [
            Interval(RealNotInt, 0, 1, closed="neither"),
            Interval(numbers.Integral, 1, None, closed="left"),
            None,
        ],
        "random_state": ["random_state"],
        "shuffle": ["boolean"],
        "stratify": ["array-like", None],
        "groups": ["array-like", None],
    },
    prefer_skip_nested_validation=True,
)
def train_test_split(
    *arrays,
    test_size: None | Float = None,
    train_size: None | Float = None,
    random_state: np.random.RandomState | None | Int = None,
    shuffle: bool = True,
    stratify: None | ArrayLike = None,
    groups: None | ArrayLike = None,
) -> list:
    """Split arrays or matrices into random train and test subsets.

    This is an extended version of ``sklearn.model_selection.train_test_split`` that
    adds support for stratified splits with non-overlapping groups. When both
    ``stratify`` and ``groups`` are provided, uses ``StratifiedGroupShuffleSplit`` to
    ensure both class distributions and group separation are preserved.

    Parameters:
        *arrays: sequence of indexables with same length / shape[0]
            Allowed inputs are lists, numpy arrays, scipy-sparse matrices or pandas
            dataframes.
        test_size: If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the test split. If int, represents the absolute
            number of test samples. If None, the value is set to the complement of the
            train size. If ``train_size`` is also None, it will be set to 0.25.
        train_size: If float, should be between 0.0 and 1.0 and represent the proportion
            of the dataset to include in the train split. If int, represents the
            absolute number of train samples. If None, the value is automatically set to
            the complement of the test size.
        random_state: Controls the randomness of the training and testing indices. Pass
            an int for reproducible output across multiple function calls.
            See :term:`Glossary <random_state>`.
        shuffle: Whether or not to shuffle the data before splitting. If False, stratify
            must be None.
        stratify: If not None, data is split in a stratified fashion, using this as the
            class labels. For binary or multiclass classification, this ensures that the
            test and training sets have approximately the same percentage of samples of
            each target class as the complete set.
        groups: Group labels for the samples used while splitting the dataset into train
            and test set. When provided with ``stratify``, ensures both stratification
            and non-overlapping groups are maintained.

    Returns:
        splitting: List containing train-test split of inputs. If ``shuffle=False``, the
            ``train`` arrays will have shape ``[0:split_point]`` and ``test`` arrays
            will have shape ``[split_point:n_samples]`` for each input.

    Examples:
        >>> import numpy as np
        >>> from ratiopath.model_selection import train_test_split
        >>> X = np.array([[1, 2], [3, 4], [5, 6], [7, 8]])
        >>> y = np.array([0, 0, 1, 1])
        >>> groups = np.array([1, 1, 2, 2])
        >>> X_train, X_test, y_train, y_test = train_test_split(
        ...     X, y, test_size=0.25, random_state=42, stratify=y, groups=groups
        ... )
        >>> X_train
        array([[1, 2],
            [5, 6],
            [7, 8]])
        >>> X_test
        array([[3, 4]])

    Notes:
        When ``shuffle=True`` and both ``stratify`` and ``groups`` are provided, uses
        ``StratifiedGroupShuffleSplit`` to split the data, ensuring that:

        * The class distribution is preserved in train and test sets
        * No group appears in both train and test sets

        When only one of ``stratify`` or ``groups`` is provided, uses the appropriate
        single-constraint splitter.

        When ``shuffle=False``, a stratified split is not supported and ``stratify``
        must be None.
    """
    n_arrays = len(arrays)
    if n_arrays == 0:
        raise ValueError("At least one array required as input")

    arrays = indexable(*arrays)

    n_samples = _num_samples(arrays[0])
    n_train, n_test = _validate_shuffle_split(
        n_samples, test_size, train_size, default_test_size=0.25
    )

    if shuffle is False:
        if stratify is not None or groups is not None:
            raise ValueError(
                "Stratified or grouped train/test split is not implemented for shuffle=False"
            )

        train = np.arange(n_train)
        test = np.arange(n_train, n_train + n_test)

    else:
        # Just this branch is different from sklearn's implementation
        if groups is not None:
            if stratify is not None:
                cvclass = StratifiedGroupShuffleSplit
            else:
                cvclass = GroupShuffleSplit
        else:
            cvclass = StratifiedShuffleSplit if stratify is not None else ShuffleSplit

        # It is safer to pass fractions, because some splitters calculate n_samplers
        # as number of groups, not samples
        cv = cvclass(
            n_splits=1,
            test_size=n_test / n_samples,
            train_size=n_train / n_samples,
            random_state=random_state,
        )

        train, test = next(cv.split(X=arrays[0], y=stratify, groups=groups))

    xp, _, device = get_namespace_and_device(arrays[0])
    train, test = move_to(train, test, xp=xp, device=device)

    return list(
        chain.from_iterable(
            (_safe_indexing(a, train), _safe_indexing(a, test)) for a in arrays
        )
    )



from collections.abc import Iterable, Iterator

import numpy as np
from numpy.typing import NDArray
from shapely import Polygon, STRtree, transform
from shapely.geometry.base import BaseGeometry


def shift_roi(
    roi: BaseGeometry, coordinate: NDArray[np.int64], downsample: float
) -> BaseGeometry:
    """Shift the region of interest (ROI) for a specific tile.

    Args:
        roi: The current region of interest geometry.
        coordinate: The coordinates of the tile to shift the ROI for.
        downsample: The downsampling factor.

    Returns:
        BaseGeometry: The shifted region of interest geometry.
    """
    return transform(roi, lambda r: r + coordinate * downsample)


def annotations_intersection(tree: STRtree, roi: BaseGeometry) -> BaseGeometry:
    """Get the intersection of annotations with the region of interest (ROI).

    Args:
        tree: The spatial index of annotation geometries.
        roi: The region of interest geometry.

    Returns:
        BaseGeometry: The intersection of annotations with the ROI.
    """
    intersection = Polygon()
    for polygon_index in tree.query(roi, predicate="intersects"):
        intersection = intersection.union(
            tree.geometries[int(polygon_index)].intersection(roi)
        )
    return intersection


def tile_annotations(
    annotations: Iterable[BaseGeometry],
    roi: BaseGeometry,
    coordinates: Iterable[NDArray[np.int64]],
    downsample: float,
) -> Iterator[BaseGeometry]:
    """Yield annotated tiles from the annotation tree.

    Annotations are assumed to be at level 0. Yielded geometries are transformed to the
    same level as the ROI and tiles.

    Args:
        annotations: The list of annotation geometries.
        roi: The region of interest geometry.
        coordinates: The iterable of coordinates of the tiles.
        downsample: The downsampling factor.

    Yields:
        BaseGeometry: The annotated tile geometry.
    """
    tree = STRtree(annotations)
    # transform roi to level 0
    roi = transform(roi, lambda geom: geom * downsample)

    for coordinate in coordinates:
        shifted_roi = shift_roi(roi, coordinate, downsample)
        intersection = annotations_intersection(tree, shifted_roi)
        yield transform(intersection, lambda geom: geom / downsample)



from collections.abc import Callable
from typing import Any

import numpy as np
from albumentations import ImageOnlyTransform
from skimage.color import combine_stains, separate_stains


class StainAugmentor(ImageOnlyTransform):
    """Applies stain augmentation to histopathological images.

    Reference:
        Tellez, D., Balkenhol, M., Karssemeijer, N., Litjens, G.,
        van der Laak, J., & Ciompi, F. (2018, March). H&E stain augmentation improves
        generalization of convolutional networks for histopathological mitosis detection.
        In Medical Imaging 2018: Digital Pathology (Vol. 10581, pp. 264-270). SPIE.
        https://geertlitjens.nl/publication/tell-18-a/tell-18-a.pdf
    """

    def __init__(
        self,
        conv_matrix: Callable[[np.ndarray], np.ndarray] | np.ndarray,
        alpha: float = 0.02,
        beta: float = 0.02,
        **kwargs: Any,
    ) -> None:
        """Initializes StainAugmentor.

        Args:
            conv_matrix (Callable[[np.ndarray], np.ndarray] | np.ndarray): Stain matrix for stain separation.
                Can be a fixed matrix or a callable that returns a matrix from an image.
            alpha (float): Multiplicative factor range for stain augmentation.
            beta (float): Additive factor range for stain augmentation.
            **kwargs (Any): Keyword arguments for ImageOnlyTransform.
        """
        super().__init__(**kwargs)
        self.conv_matrix = conv_matrix
        self.alpha = alpha
        self.beta = beta

        if isinstance(self.conv_matrix, np.ndarray):
            self.inv_conv_matrix = np.linalg.inv(self.conv_matrix)

    def apply(
        self,
        img: np.ndarray,
        conv_matrix: np.ndarray,
        inv_conv_matrix: np.ndarray,
        alphas: list[float],
        betas: list[float],
        **params: dict[str, Any],
    ) -> np.ndarray:
        stains = separate_stains(img, inv_conv_matrix)

        for i in range(stains.shape[-1]):
            stains[..., i] *= alphas[i]
            stains[..., i] += betas[i]

        return np.astype(combine_stains(stains, conv_matrix) * 255, np.uint8)

    def get_params_dependent_on_data(
        self, params: dict[str, Any], data: dict[str, Any]
    ) -> dict[str, Any]:
        conv_matrix = (
            self.conv_matrix(data["image"])
            if callable(self.conv_matrix)
            else self.conv_matrix
        )
        inv_conv_matrix = (
            self.inv_conv_matrix
            if hasattr(self, "inv_conv_matrix")
            else np.linalg.inv(conv_matrix)
        )

        return {
            "conv_matrix": conv_matrix,
            "inv_conv_matrix": inv_conv_matrix,
            "alphas": [
                self.py_random.uniform(1 - self.alpha, 1 + self.alpha)
                for _ in range(conv_matrix.shape[-1])
            ],
            "betas": [
                self.py_random.uniform(-self.beta, self.beta)
                for _ in range(conv_matrix.shape[-1])
            ],
        }



import pyvips

from ratiopath.masks.vips_filters.typed import Res
from ratiopath.masks.vips_filters.vips_filter import VipsFilter


# The default value 10 is based on empirical observations.
DISC_FACTOR = 10


class VipsMorphology(VipsFilter):
    """Base class for morphological operations using disk structuring elements."""

    def __init__(self, disc_factor: int = DISC_FACTOR) -> None:
        self.disc_factor = disc_factor

    def _disc_size(self, mpp: Res) -> int:
        return round(self.disc_factor / (sum(mpp) / len(mpp)))

    def _disc_object(self, mpp: Res) -> pyvips.Image:
        disc_size = self._disc_size(mpp)

        return (
            pyvips.Image.black(2 * disc_size + 1, 2 * disc_size + 1) + 128
        ).draw_circle(255, disc_size, disc_size, disc_size, fill=True)


class VipsOpening(VipsMorphology):
    """Applies an opening operation to an image using a disk structuring element."""

    def __call__(self, image: pyvips.Image, mpp: Res) -> tuple[pyvips.Image, Res]:
        disc = self._disc_object(mpp)

        image = image.morph(disc, pyvips.enums.OperationMorphology.ERODE)
        image = image.morph(disc, pyvips.enums.OperationMorphology.DILATE)
        return image, mpp


class VipsClosing(VipsMorphology):
    """Applies a closing operation to an image using a disk structuring element."""

    def __call__(self, image: pyvips.Image, mpp: Res) -> tuple[pyvips.Image, Res]:
        disc = self._disc_object(mpp)

        image = image.morph(disc, pyvips.enums.OperationMorphology.DILATE)
        image = image.morph(disc, pyvips.enums.OperationMorphology.ERODE)
        return image, mpp



import os
from pathlib import Path
from typing import TypeVar

import pandas as pd
from PIL import Image

from ratiopath.openslide import OpenSlide


T = TypeVar("T")


class OpenSlideTileReader:
    def __init__(
        self,
        path: str | Path,
        level: int | str | None = None,
        resolution: float | tuple[float, float] | None = None,
        background: None | tuple[int, int, int] = (255, 255, 255),
    ) -> None:
        assert (level is None) != (resolution is None), (
            "Either level or resolution must be provided"
        )

        assert os.path.exists(path), f"Path {path} does not exist"

        self.path = Path(path)
        self.background = background

        # Assign level based on the provided arguments
        if resolution is not None:
            with OpenSlide(self.path) as slide:
                self.level: int | str = slide.closest_level(resolution)
        elif level is not None:
            self.level = level

    def get_openslide_tile(
        self,
        tile_coords: tuple[int, int],
        tile_extent: tuple[int, int],
        tile: pd.Series,
    ) -> Image.Image:
        """Returns tile from the slide image at the specified coordinates in RGB format."""
        level = self._get_from_tile(tile, self.level)

        with OpenSlide(self.path) as slide:
            rgba_region = slide.read_region_relative(tile_coords, level, tile_extent)

            if self.background is not None:
                # Create a new image with the background color
                background = Image.new("RGBA", rgba_region.size, self.background)
                # Composite the region with the background
                rgba_region = Image.alpha_composite(background, rgba_region)

        return rgba_region.convert("RGB")

    def _get_from_tile(self, tile: pd.Series, key: T | str) -> T:
        return tile[key] if isinstance(key, str) else key



from abc import ABC, abstractmethod
from collections.abc import Iterable

import pyvips

from ratiopath.masks.vips_filters.typed import Res


class VipsFilter(ABC):
    @abstractmethod
    def __call__(self, image: pyvips.Image, mpp: Res) -> tuple[pyvips.Image, Res]: ...


class VipsCompose(VipsFilter):
    def __init__(self, filters: Iterable[VipsFilter]) -> None:
        self.filters = filters

    def __call__(self, image: pyvips.Image, mpp: Res) -> tuple[pyvips.Image, Res]:
        for filter in self.filters:
            image, mpp = filter(image, mpp)
        return image, mpp



import json
from collections.abc import Iterable
from pathlib import Path
from typing import Any, TextIO

import pandas as pd
from geopandas import GeoDataFrame
from shapely.geometry import Point, Polygon, box
from shapely.geometry.base import BaseGeometry


class Darwin7JSONParser:
    """Parser for Darwin7 JSON format annotation files.

    Extracts geometries and metadata from Darwin JSON structures into a GeoDataFrame.
    Supports Polygon, Point, and bounding box geometries.
    Expected JSON schema:
    Root
    └── annotations (list)
        ├── Feature (Polygon)
        │   ├── id: "uuid"
        │   ├── name: "class_name"
        │   ├── properties: [...] or {...}
        │   ├── slot_names: [...]
        │   └── polygon
        │       └── paths: [ [ {x, y}, ... ], [holes...] ]
        ├── Feature (Bounding Box)
        │   ├── id: "uuid"
        │   └── bounding_box: {x, y, w, h}
        └── Feature (Point)
            ├── id: "uuid"
            └── point: {x, y}
    """

    def __init__(self, file_path: Path | str | TextIO) -> None:
        if isinstance(file_path, (str, Path)):
            with open(file_path, encoding="utf-8") as f:
                data = json.load(f)
        else:
            data = json.load(file_path)

        records = []
        for ann in data.get("annotations", []):
            props = ann.get("properties", [])
            records.append(
                {
                    "id": ann.get("id"),
                    "name": ann.get("name"),
                    "properties": json.dumps(props)
                    if isinstance(props, (list, dict))
                    else props,
                    "slot_names": json.dumps(ann.get("slot_names", [])),
                    "geometry": self._parse_geometry(ann),
                }
            )

        if records:
            self.gdf = GeoDataFrame(pd.DataFrame(records), geometry="geometry")
        else:
            self.gdf = GeoDataFrame(
                columns=["id", "name", "properties", "slot_names", "geometry"],
                geometry="geometry",
            )

    @staticmethod
    def _parse_geometry(ann: dict[str, Any]) -> BaseGeometry | None:
        """Construct Shapely geometry objects from Darwin annotation dictionaries."""
        if "polygon" in ann:
            paths = ann["polygon"].get("paths", [])
            if paths:
                exterior = [(pt["x"], pt["y"]) for pt in paths[0]]
                holes = [
                    [(pt["x"], pt["y"]) for pt in hole]
                    for hole in paths[1:]
                    if len(hole) >= 3
                ]
                if len(exterior) >= 3:
                    return Polygon(exterior, holes)
        elif "bounding_box" in ann:
            bbox = ann["bounding_box"]
            return box(
                bbox["x"], bbox["y"], bbox["x"] + bbox["w"], bbox["y"] + bbox["h"]
            )
        elif "point" in ann:
            return Point(ann["point"]["x"], ann["point"]["y"])

        return None

    @staticmethod
    def extract_nested(val: Any, path: list[str]) -> Any:
        """Extract a nested value from a JSON-like structure using a list of keys."""
        if isinstance(val, str):
            try:
                val = json.loads(val)
            except json.JSONDecodeError:
                return None
        for key in path:
            if isinstance(val, dict) and key in val:
                val = val[key]
            elif isinstance(val, list) and key.isdigit():
                idx = int(key)
                if idx < len(val):
                    val = val[idx]
                else:
                    return None
            else:
                return None
        return val

    def get_filtered_geodataframe(
        self, separator: str = "_", **kwargs: str
    ) -> GeoDataFrame:
        """Filter the GeoDataFrame based on property values.

        Supports filtering by top-level columns or nested attributes within JSON-like
        columns (e.g., 'properties'). Nested keys are accessed by joining column
        names and keys with the separator.

        Args:
            separator: String used to separate nested keys in kwargs.
            **kwargs: Keyword arguments where keys represent (possibly nested)
                columns and values are regex patterns to match.

        Returns:
            A GeoDataFrame containing only the rows that match all filter criteria.
            If a requested top-level column is missing, an empty GeoDataFrame
            with the original schema is returned.
        """
        filtered_gdf = self.gdf

        for key, pattern in kwargs.items():
            subkeys = key.split(separator)
            if not subkeys or subkeys[0] not in filtered_gdf.columns:
                return self.gdf.iloc[0:0]

            series = filtered_gdf[subkeys[0]]

            if len(subkeys) > 1:
                series = series.apply(
                    lambda x, sk=subkeys[1:]: self.extract_nested(x, sk)
                )
                mask = series.notna()
                filtered_gdf = filtered_gdf[mask]
                series = series[mask]

            if filtered_gdf.empty:
                return filtered_gdf

            series = series.astype(str)
            mask = series.str.match(pattern, na=False)
            filtered_gdf = filtered_gdf[mask]

        return filtered_gdf

    def get_polygons(self, **kwargs: str) -> Iterable[Polygon]:
        """Get polygons from the GeoDataFrame.

        Args:
            **kwargs: Keyword arguments containing regex patterns for filtering properties.

        Yields:
            Shapely Polygon objects.
        """
        filtered_gdf = self.get_filtered_geodataframe(**kwargs)
        yield from filtered_gdf[filtered_gdf.geom_type == "Polygon"].geometry

    def get_points(self, **kwargs: str) -> Iterable[Point]:
        """Get points from the GeoDataFrame.

        Args:
            **kwargs: Keyword arguments containing regex patterns for filtering properties.

        Yields:
            Shapely Point objects.
        """
        filtered_gdf = self.get_filtered_geodataframe(**kwargs)
        yield from filtered_gdf[filtered_gdf.geom_type == "Point"].geometry



from ratiopath.augmentations.estimate_stain_vectors import estimate_stain_vectors
from ratiopath.augmentations.stain_augmentor import StainAugmentor


__all__ = ["StainAugmentor", "estimate_stain_vectors"]



"""Tests for annotation processing functions."""

import io
import json
from math import isclose

import numpy as np
import pytest
from shapely import Polygon

from ratiopath.parsers import GeoJSONParser
from ratiopath.tiling.annotations import tile_annotations


class TestMapAnnotations:
    """Test the map_annotations function."""

    @pytest.fixture
    def sample_geojson_content(self):
        """Sample GeoJSON content for testing."""
        return {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "geometry": {
                        "type": "Polygon",
                        "coordinates": [[[0, 0], [10, 0], [10, 10], [0, 10], [0, 0]]],
                    },
                    "properties": {},
                },
                {
                    "type": "Feature",
                    "geometry": {"type": "Point", "coordinates": [5, 5]},
                    "properties": {},
                },
            ],
        }

    def test_map_annotations_basic(self, sample_geojson_content):
        """Test basic functionality of map_annotations."""
        # Create a temporary annotation file
        f = io.StringIO(json.dumps(sample_geojson_content))

        results = tile_annotations(
            annotations=list(GeoJSONParser(f).get_polygons()),
            roi=Polygon([(0, 0), (8, 0), (8, 8), (0, 8)]),
            coordinates=np.array([[0, 0], [8, 0], [0, 8], [8, 8]]),
            downsample=1,
        )

        for result_polygon, area in zip(results, [64.0, 16.0, 16.0, 4.0], strict=True):
            assert result_polygon.is_valid
            assert isclose(result_polygon.area, area)

    def test_map_annotations_custom_roi(self, sample_geojson_content):
        """Test map_annotations with a custom region of interest."""
        f = io.StringIO(json.dumps(sample_geojson_content))

        results = tile_annotations(
            annotations=list(GeoJSONParser(f).get_polygons()),
            roi=Polygon([(1, 1), (7, 1), (7, 7), (1, 7)]),
            coordinates=np.array([[0, 0], [8, 0], [0, 8], [8, 8]]),
            downsample=1,
        )

        for result_polygon, area in zip(results, [36.0, 6.0, 6.0, 1.0], strict=True):
            assert result_polygon.is_valid
            assert isclose(result_polygon.area, area)



from collections.abc import Iterator
from sys import getsizeof
from typing import Any

import numpy as np
import pyarrow

from ray.data.block import Block, BlockMetadata
from ray.data.datasource import FileBasedDatasource
from ray.data.datasource.file_meta_provider import DefaultFileMetadataProvider


class SlideMetaDatasource(FileBasedDatasource):
    """A Ray datasource that extracts metadata from whole slide image files.

    This datasource reads whole slide image files and extracts metadata required for
    tiled processing. For each slide, it determines the best level to use based on
    the provided `mpp` (microns per pixel) or `level`, and extracts properties
    like dimensions, MPP, and downsampling factor for that level.

    The output of this datasource is a Ray Dataset where each row corresponds to a
    single slide and contains the necessary metadata for subsequent tiling operations.

    Supported file formats:
        - OpenSlide: `svs`, `tif`, `dcm`, `ndpi`, `vms`, `vmu`, `scn`, `mrxs`, `tiff`, `svslide`, `bif`, `czi`
        - OME-TIFF: `ome.tiff`, `ome.tif`

    Output columns:
        - path (str): Path to the slide file.
        - extent_x (int): Width of the selected slide level in pixels.
        - extent_y (int): Height of the selected slide level in pixels.
        - tile_extent_x (int): Width of the tiles to be extracted.
        - tile_extent_y (int): Height of the tiles to be extracted.
        - stride_x (int): Horizontal stride for tiling.
        - stride_y (int): Vertical stride for tiling.
        - mpp_x (float): Microns per pixel in the x-direction for the selected level.
        - mpp_y (float): Microns per pixel in the y-direction for the selected level.
        - level (int): The selected slide level.
        - downsample (float): The downsample factor for the selected level.
    """

    def __init__(
        self,
        paths: str | list[str],
        *,
        mpp: float | None = None,
        level: int | None = None,
        tile_extent: int | tuple[int, int],
        stride: int | tuple[int, int],
        **file_based_datasource_kwargs: Any,
    ) -> None:
        """Initializes the SlideMetaDatasource.

        Args:
            paths: A path or list of paths to whole slide image files.
            mpp: The desired microns per pixel. The datasource will select the slide
                level with the closest MPP. Exactly one of `mpp` or `level` must be
                provided.
            level: The desired slide level to use. Exactly one of `mpp` or `level`
                must be provided.
            tile_extent: The size of the tiles to be generated, as (width, height).
                If a single integer is provided, it's used for both dimensions.
            stride: The step size between consecutive tiles, as (x_stride, y_stride).
                If a single integer is provided, it's used for both dimensions.
            **file_based_datasource_kwargs: Additional keyword arguments passed to
                the base `FileBasedDatasource`.
        """
        super().__init__(paths, **file_based_datasource_kwargs)

        assert (mpp is not None) != (level is not None), (
            "Exactly one of 'mpp' or 'level' must be provided, not both or neither."
        )

        self.desired_mpp = mpp
        self.desired_level = level
        self.tile_extent = np.broadcast_to(tile_extent, 2)
        self.stride = np.broadcast_to(stride, 2)

    def _read_stream(self, f: pyarrow.NativeFile, path: str) -> Iterator[Block]:
        if path.lower().endswith((".ome.tiff", ".ome.tif")):
            return self._read_ome_stream(f, path)
        return self._read_openslide_stream(f, path)

    def _read_ome_stream(self, f: pyarrow.NativeFile, path: str) -> Iterator[Block]:
        from ratiopath.tifffile import TiffFile

        with TiffFile(path) as tif:
            if self.desired_level is not None:
                level = self.desired_level
            else:
                assert self.desired_mpp is not None
                level = tif.closest_level(self.desired_mpp)

            mpp = tif.slide_resolution(level)

            extent_y, extent_x, *_ = tif.get_main_page(level).shape
            downsample = tif.get_main_page(0).shape[1] / extent_x

        yield self._build_block(path, (extent_x, extent_y), mpp, level, downsample)

    def _read_openslide_stream(
        self, f: pyarrow.NativeFile, path: str
    ) -> Iterator[Block]:
        from ratiopath.openslide import OpenSlide

        with OpenSlide(path) as slide:
            if self.desired_level is not None:
                level = self.desired_level
            else:
                assert self.desired_mpp is not None
                level = slide.closest_level(self.desired_mpp)
            mpp_x, mpp_y = slide.slide_resolution(level)

            extent = slide.level_dimensions[level]
            downsample = slide.level_downsamples[level]

        yield self._build_block(path, extent, (mpp_x, mpp_y), level, downsample)

    def _build_block(
        self,
        path: str,
        extent: tuple[int, int],
        mpp: tuple[float, float],
        level: int,
        downsample: float,
    ) -> Block:
        from ray.data._internal.delegating_block_builder import DelegatingBlockBuilder

        builder = DelegatingBlockBuilder()
        item = {
            "path": path,
            "extent_x": extent[0],
            "extent_y": extent[1],
            "tile_extent_x": self.tile_extent[0],
            "tile_extent_y": self.tile_extent[1],
            "stride_x": self.stride[0],
            "stride_y": self.stride[1],
            "mpp_x": mpp[0],
            "mpp_y": mpp[1],
            "level": level,
            "downsample": downsample,
        }
        builder.add(item)
        return builder.build()

    def _rows_per_file(self) -> int:
        return 1

    def estimate_inmemory_data_size(self) -> int | None:
        paths = self._paths()
        if not paths:
            return 0

        # Create a sample item to calculate the base size of a single row.
        sample_item = {
            "path": "",
            "extent_x": 0,
            "extent_y": 0,
            "tile_extent_x": 0,
            "tile_extent_y": 0,
            "stride_x": 0,
            "stride_y": 0,
            "mpp_x": 0.0,
            "mpp_y": 0.0,
            "level": 0,
            "downsample": 0.0,
        }

        # Calculate the size of the dictionary structure, keys, and fixed-size values.
        base_row_size = getsizeof(sample_item)
        for k, v in sample_item.items():
            base_row_size += getsizeof(k)
            base_row_size += getsizeof(v)

        # Calculate the total size of all path strings.
        total_path_size = sum(getsizeof(p) for p in paths)

        # The total estimated size is the base size for each row plus the total size of paths.
        return base_row_size * len(paths) + total_path_size


class SlideFileMetadataProvider(DefaultFileMetadataProvider):
    def _get_block_metadata(  # type: ignore[override]
        self,
        paths: list[str],
        *,
        rows_per_file: int | None,
        file_sizes: list[int | None],
    ) -> BlockMetadata:
        sample_item = {
            "path": "",
            "extent_x": 0,
            "extent_y": 0,
            "tile_extent_x": 0,
            "tile_extent_y": 0,
            "stride_x": 0,
            "stride_y": 0,
            "mpp_x": 0.0,
            "mpp_y": 0.0,
            "level": 0,
            "downsample": 0.0,
        }
        base_row_size = getsizeof(sample_item)
        for k, v in sample_item.items():
            base_row_size += getsizeof(k)
            base_row_size += getsizeof(v)

        # Estimate size for each path and create a new list of file sizes.
        estimated_file_sizes = [base_row_size + getsizeof(p) for p in paths]

        return super()._get_block_metadata(
            paths,
            rows_per_file=rows_per_file,
            file_sizes=estimated_file_sizes,  # type: ignore[arg-type]
        )



from typing import Self, cast

import numpy as np
import tifffile


class TiffFile(tifffile.TiffFile):  # type: ignore [misc]
    """A wrapper around the TiffFile library to provide additional functionality."""

    def __enter__(self) -> Self:
        """Handles entering the context manager."""
        super().__enter__()
        return self

    def levels(self) -> int:
        """Returns the number of levels in the OME-TIFF file.

        Returns:
            The number of levels in the OME-TIFF file.
        """
        return len(self.series[0].pages)

    def get_main_page(self, level: int) -> "tifffile.TiffPage":
        """Reads the desired level (page) at the series of index 0. It usually refers to the main scene.

        Args:
            level: The level of the OME-TIFF file.

        Returns:
            The OME-TIFF page of series at index 0 at desired level.
        """
        return cast("tifffile.TiffPage", self.series[0].pages[level])

    def slide_resolution(self, level: int) -> tuple[float, float]:
        """Returns the physical resolution (µm/px) of the OME-TIFF file at the given level.

        Args:
            level: The level of the OME-TIFF file to get the MPP.

        Returns:
            The (x, y) physical resolution of the OME-TIFF file in µm/px.
        """
        page = self.get_main_page(level)

        # Tag 282/283 are Rational (Num, Denom)
        x_res: tuple[int, int] = page.tags["XResolution"].value
        y_res: tuple[int, int] = page.tags["YResolution"].value
        # Tag 296 is ResolutionUnit
        unit: tifffile.RESUNIT = page.tags["ResolutionUnit"].value

        # Resolve value (numerator / denominator)
        x_val = x_res[0] / x_res[1]
        y_val = y_res[0] / y_res[1]

        match unit:
            case tifffile.RESUNIT.INCH:  # Inch
                return (25400 / x_val, 25400 / y_val)
            case tifffile.RESUNIT.CENTIMETER:  # Centimeter
                return (10000 / x_val, 10000 / y_val)
            case tifffile.RESUNIT.MILLIMETER:  # Milimeter
                return (1000 / x_val, 1000 / y_val)
            case tifffile.RESUNIT.MICROMETER:  # Micrometer
                return (x_val, y_val)
            case _:
                raise ValueError(f"Unsupported ResolutionUnit: {unit}")

    def closest_level(self, mpp: float | tuple[float, float]) -> int:
        """Finds the closest OME-TIFF level to match the desired MPP.

        This method compares the desired MPP (µm/px) with the MPP of the
        available levels in the OME-TIFF file and selects the level with the closest match.

        Args:
            mpp: The desired µm/px value.

        Returns:
            The index of the level with the closest µm/px resolution to the desired value.
        """
        available_levels_mpp = [
            self.slide_resolution(level) for level in range(self.levels())
        ]

        return (
            np.abs(np.asarray(available_levels_mpp) - np.asarray(mpp))
            .mean(axis=1)
            .argmin()
            .item()
        )



import numpy as np

from ratiopath.model_selection.split import (
    StratifiedGroupShuffleSplit,
    train_test_split,
)


def test_train_test_split_with_groups_and_stratify():
    x = np.arange(12).reshape(6, 2)
    y = np.array([0, 0, 1, 1, 0, 1])
    groups = np.array([1, 1, 2, 2, 3, 3])

    # include groups as one of the arrays so we can inspect split groups
    _, _, _, y_test, g_train, g_test = train_test_split(
        x, y, groups, test_size=0.33, random_state=0, stratify=y, groups=groups
    )

    # ensure groups do not overlap between train and test
    assert set(g_train).isdisjoint(set(g_test))

    # ensure stratification roughly preserved in the test set
    prop_full = (y == 0).sum() / len(y)
    prop_test = (y_test == 0).sum() / len(y_test)
    assert abs(prop_full - prop_test) <= 0.34


def test_train_test_split_with_groups_no_stratify():
    x = np.arange(10).reshape(5, 2)
    y = np.array([0, 1, 0, 1, 0])
    groups = np.array([1, 1, 2, 2, 3])

    _, _, _, _, g_train, g_test = train_test_split(
        x, y, groups, test_size=0.4, random_state=1, groups=groups
    )

    assert set(g_train).isdisjoint(set(g_test))


def test_stratified_group_shuffle_split_splits():
    x = np.arange(12).reshape(6, 2)
    y = np.array([0, 0, 1, 1, 0, 1])
    groups = np.array([1, 1, 2, 2, 3, 3])

    sgss = StratifiedGroupShuffleSplit(n_splits=5, test_size=0.33, random_state=42)

    for train_idx, test_idx in sgss.split(x, y, groups=groups):
        # groups should be non-overlapping
        train_groups = set(groups[train_idx])
        test_groups = set(groups[test_idx])
        assert train_groups.isdisjoint(test_groups)

        # indices should cover all samples
        assert len(train_idx) + len(test_idx) == len(x)

        # test must contain at least one sample
        assert len(test_idx) > 0



from typing import Any

import pyarrow

from ray.data.datasource.file_datasink import RowBasedFileDatasink


class VipsTiffDatasink(RowBasedFileDatasink):
    """A datasink for saving image data as TIFF files using libvips.

    This datasink uses pyvips to efficiently save image data with support for
    various TIFF formats, including BigTIFF.
    """

    def __init__(
        self,
        path: str,
        data_column: str,
        options_column: str | None = None,
        default_options: dict[str, Any] | None = None,
        **file_datasink_kwargs: Any,
    ) -> None:
        """Initialize a VipsTIFFDatasink.

        Args:
            path: Output path for the TIFF files
            data_column: Column name containing image data (numpy array)
            options_column: Optional column name containing TIFF save options as dict
            default_options: Default options for TIFF saving
            **file_datasink_kwargs: Additional arguments for the file datasink
        """
        super().__init__(path, file_format="tiff", **file_datasink_kwargs)
        self.data_column = data_column
        self.options_column = options_column
        self.default_options = default_options or {}

    def write_row_to_file(self, row: dict[str, Any], file: pyarrow.NativeFile) -> None:
        from pyvips import Image

        image = Image.new_from_array(row[self.data_column])
        options = self.default_options.copy()
        if self.options_column is not None:
            options.update(row.get(self.options_column, {}))

        buffer = image.tiffsave_buffer(**options)
        file.write(buffer)



from typing import Unpack

import numpy as np
import pyarrow as pa
from ray.data.datatype import DataType
from ray.data.expressions import udf
from ray.data.extensions import TensorArray

from ratiopath.openslide import OpenSlide
from ratiopath.tifffile import TiffFile
from ratiopath.tiling.types import ReadTilesArguments
from ratiopath.tiling.utils import (
    _pyarrow_group_indices,
    _read_openslide_tiles,
    _read_tifffile_tiles,
)


def read_openslide_tiles(path: str, **kwargs: Unpack[ReadTilesArguments]) -> np.ndarray:
    """Read batch of tiles from a whole-slide image using OpenSlide."""
    with OpenSlide(path) as slide:
        return _read_openslide_tiles(slide, **kwargs)


def read_tifffile_tiles(path: str, **kwargs: Unpack[ReadTilesArguments]) -> np.ndarray:
    """Read batch of tiles from an OME-TIFF file using tifffile."""
    with TiffFile(path) as slide:
        return _read_tifffile_tiles(slide, **kwargs)


@udf(return_dtype=DataType(np.ndarray))
def read_slide_tiles(
    path: pa.StringArray,
    tile_x: pa.IntegerArray,
    tile_y: pa.IntegerArray,
    tile_extent_x: pa.IntegerArray,
    tile_extent_y: pa.IntegerArray,
    level: pa.IntegerArray,
) -> pa.Array:
    """Reads a batch of tiles from a whole-slide image using either OpenSlide or tifffile.

    Args:
        path: A pyarrow array of whole-slide image paths.
        tile_x: A pyarrow array of tile x-coordinates.
        tile_y: A pyarrow array of tile y-coordinates.
        tile_extent_x: A pyarrow array of tile extents in the x-dimension.
        tile_extent_y: A pyarrow array of tile extents in the y-dimension.
        level: A pyarrow array of slide levels to read the tiles from.

    Returns:
        A pyarrow array of numpy arrays containing the read tiles.
    """
    import pyarrow.compute as pc

    tiles = np.empty(len(tile_x), dtype=object)

    for p, group in _pyarrow_group_indices(path).items():
        assert isinstance(p, str)

        kwargs = {
            "tile_x": pc.take(tile_x, group),
            "tile_y": pc.take(tile_y, group),
            "tile_extent_x": pc.take(tile_extent_x, group),
            "tile_extent_y": pc.take(tile_extent_y, group),
            "level": pc.take(level, group),
        }

        # Check if it's an OME-TIFF file
        if p.lower().endswith((".ome.tiff", ".ome.tif")):
            tiles[group] = list(read_tifffile_tiles(p, **kwargs))
        else:
            tiles[group] = list(read_openslide_tiles(p, **kwargs))

    return pa.array(TensorArray(tiles))



from ratiopath.tiling.annotations import tile_annotations
from ratiopath.tiling.overlays import tile_overlay, tile_overlay_overlap
from ratiopath.tiling.read_slide_tiles import read_slide_tiles
from ratiopath.tiling.tilers import grid_tiles


__all__ = [
    "grid_tiles",
    "read_slide_tiles",
    "tile_annotations",
    "tile_overlay",
    "tile_overlay_overlap",
]



from typing import TypedDict

import pyarrow as pa


class ReadTilesArguments(TypedDict):
    """TypedDict for read_tiles arguments."""

    tile_x: pa.IntegerArray
    tile_y: pa.IntegerArray
    tile_extent_x: pa.IntegerArray
    tile_extent_y: pa.IntegerArray
    level: pa.IntegerArray


class ReadOverlaysArguments(TypedDict):
    """TypedDict for read_overlays arguments."""

    tile_x: pa.IntegerArray
    tile_y: pa.IntegerArray
    tile_extent_x: pa.IntegerArray
    tile_extent_y: pa.IntegerArray
    mpp_x: pa.FloatArray
    mpp_y: pa.FloatArray



import json
from collections.abc import Iterable
from pathlib import Path
from typing import TextIO

import geopandas as gpd
from geopandas import GeoDataFrame
from shapely import Point, Polygon


class GeoJSONParser:
    """Parser for GeoJSON format annotation files.

    GeoJSON is a format for encoding geographic data structures using JSON.
    This parser supports both polygon and point geometries.
    """

    def __init__(self, file_path: Path | str | TextIO) -> None:
        self.gdf = gpd.read_file(file_path)

        if not self.gdf.empty:
            # Explode Multi-part geometries to simplify geometry handling
            self.gdf = self.gdf.explode(index_parts=True)

    def get_filtered_geodataframe(
        self, separator: str = "_", **kwargs: str
    ) -> GeoDataFrame:
        """Filter the GeoDataFrame based on property values.

        Args:
            separator: The string used to separate keys in the filtering.
            **kwargs: Keyword arguments for filtering. Keys are column names
                (e.g., 'classification.name') and values are regex patterns to match
                against.

        Returns:
            A filtered GeoDataFrame.
        """
        filtered_gdf = self.gdf
        for key, pattern in kwargs.items():
            subkeys = key.split(separator)
            if not subkeys or subkeys[0] not in filtered_gdf.columns:
                # If the first part of the key doesn't exist, return an empty frame
                return self.gdf.iloc[0:0]

            series = filtered_gdf[subkeys[0]].astype(str)
            if len(subkeys) > 1:
                mask = series.apply(is_json_dict)
                series = series[mask].apply(lambda x: json.loads(x))
                filtered_gdf = filtered_gdf[mask]

            for subkey in subkeys[1:]:
                mask = series.apply(
                    lambda x, subkey=subkey: isinstance(x, dict) and subkey in x
                )
                series = series[mask].apply(lambda x, subkey=subkey: x[subkey])
                filtered_gdf = filtered_gdf[mask]

            series = series.astype(str)
            mask = series.str.match(pattern, na=False)
            filtered_gdf = filtered_gdf[mask]

        return filtered_gdf

    def get_polygons(self, **kwargs: str) -> Iterable[Polygon]:
        """Get polygons from the GeoDataFrame.

        Args:
            **kwargs: Keyword arguments for filtering properties.

        Yields:
            Shapely Polygon objects.
        """
        filtered_gdf = self.get_filtered_geodataframe(**kwargs)
        for geom in filtered_gdf.geometry:
            if isinstance(geom, Polygon):
                yield geom

    def get_points(self, **kwargs: str) -> Iterable[Point]:
        """Get points from the GeoDataFrame.

        Args:
            **kwargs: Keyword arguments for filtering properties.

        Yields:
            Shapely Point objects.
        """
        filtered_gdf = self.get_filtered_geodataframe(**kwargs)
        for geom in filtered_gdf.geometry:
            if isinstance(geom, Point):
                yield geom


def is_json_dict(string: str) -> bool:
    try:
        valid_json = json.loads(string)
        if isinstance(valid_json, dict):
            return True
    except ValueError:
        return False
    return False



import itertools
import warnings
from collections.abc import Iterator
from typing import Literal, TypeVarTuple

import numpy as np
from numpy.typing import NDArray


Dims = TypeVarTuple("Dims")


def grid_tiles(
    slide_extent: tuple[int, *Dims],
    tile_extent: tuple[int, *Dims],
    stride: tuple[int, *Dims],
    last: Literal["shift", "drop", "keep"] = "keep",
) -> Iterator[NDArray[np.int64]]:
    """Generates tiles for the given slide based on its size, tile size, and stride.

    The function yields tile coordinates in row-major order, iterating first over the
    x-axis (e.g. (0,0,...), (1,0,...), (2,0,...)) before incrementing the y-axis
    (0,1,...), (1,1,...), etc.

    Args:
        slide_extent: The dimensions of the slide in pixels.
        tile_extent: The dimensions of the tile in pixels.
        stride: The stride between tiles in pixels.
        last: The strategy to handle the last tile when it does not fit the stride.
            - "shift": Shift the last tile to the left and up to fit the stride.
            - "drop": Drop the last tile if it does not fit the stride.
            - "keep": Keep the last tile even if it does not fit the slide.

    Returns:
        An iterator of numpy arrays containing the tile coordinates.
    """
    slide_extent_array = np.asarray(slide_extent)
    tile_extent_array = np.asarray(tile_extent)
    stride_array = np.asarray(stride)

    if any(tile_extent_array > slide_extent_array):
        warnings.warn(
            f"TilingModule: tile size {tile_extent_array} is greater than slide dimensions {slide_extent_array}",
            UserWarning,
            stacklevel=2,
        )

    dim_max = (slide_extent_array - tile_extent_array) / stride_array
    dim_max = np.floor(dim_max) if last == "drop" else np.ceil(dim_max)
    dim_max = dim_max.astype(int)

    # Reverse the dimension max array to iterate over 'x' coordinates first
    dim_max = dim_max[::-1]

    # Generate tile coordinates
    if last == "drop" or last == "keep":
        for i in itertools.product(*map(range, dim_max + 1)):
            yield np.array(i[::-1]) * stride_array

    elif last == "shift":
        for i in itertools.product(*map(range, dim_max + 1)):
            base_coord = np.array(i[::-1]) * stride_array
            yield np.minimum(base_coord, slide_extent_array - tile_extent_array)



from ratiopath.ray.datasource.slide_metadatasource import SlideMetaDatasource
from ratiopath.ray.datasource.vips_tiff_datasink import VipsTiffDatasink


__all__ = ["SlideMetaDatasource", "VipsTiffDatasink"]



import pyvips

from ratiopath.masks.vips_filters import (
    VipsClosing,
    VipsCompose,
    VipsFilter,
    VipsGrayScaleFilter,
    VipsOpening,
    VipsOtsu,
)
from ratiopath.masks.vips_filters.typed import Res


default_filter = VipsCompose(
    [
        VipsGrayScaleFilter(),
        VipsOtsu(),
        VipsOpening(),
        VipsClosing(),
    ]
)


def tissue_mask(
    slide: pyvips.Image, mpp: Res, filter: VipsFilter = default_filter
) -> tuple[pyvips.Image, Res]:
    """Generates a tissue mask from a whole-slide image (WSI) using saturation channel extraction and morphological operations, and saves the mask as a TIFF image.

    The function extracts the saturation channel from the WSI, applies Otsu thresholding to
    identify tissue regions, and performs morphological operations (closing and opening)
    to refine the mask.

    Args:
        slide: whole-slide image (WSI) pyvips file handler.
        mpp: Resolution of the mask in µm/px.
        filter: A VipsFilter object that defines the operations to be applied to the image.
            The default filter includes saturation channel extraction, Otsu thresholding, closing, and opening operations.
            The disk factor for morphological operations using the default filter is 10.

    Returns:
        The generated tissue mask as pyvips.Image.
    """
    return filter(slide, mpp)



import hashlib
from collections.abc import Callable
from typing import Any

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc

from ratiopath.openslide import OpenSlide
from ratiopath.tifffile import TiffFile


def row_hash(
    row: dict[str, Any],
    column: str = "id",
    algorithm: Callable[[bytes], hashlib._hashlib.HASH] = hashlib.sha256,  # type: ignore[name-defined]
) -> dict[str, Any]:
    """Hashes a row (dictionary) using SHA256 and adds the hash as a new column.

    Args:
        row: The dictionary (row) to hash.
        column: The name of the column to store the hash. Defaults to "id".
        algorithm: The hashing algorithm to use. Defaults to hashlib.sha256.

    Returns:
        The modified row (dictionary) with the SHA256 hash added.
    """
    row[column] = algorithm(str(row).encode()).hexdigest()
    return row


def _read_openslide_tiles(
    slide: OpenSlide,
    tile_x: pa.IntegerArray,
    tile_y: pa.IntegerArray,
    tile_extent_x: pa.IntegerArray,
    tile_extent_y: pa.IntegerArray,
    level: pa.IntegerArray,
    background: int = 255,
) -> np.ndarray:
    """Read batch of tiles from a whole-slide image using OpenSlide.

    Args:
        slide: The OpenSlide object to read tiles from.
        tile_x: Array of tile x-coordinates (in pixels) at level 0.
        tile_y: Array of tile y-coordinates (in pixels) at level 0.
        tile_extent_x: Array of tile widths in pixels.
        tile_extent_y: Array of tile heights in pixels.
        level: Array of OpenSlide level indices for each tile.
        background: The RGB value (0-255) to use for transparent areas.
            Defaults to 255 (white).

    Returns:
        A NumPy array of RGB tiles read from the slide.
    """
    from PIL import Image

    def get_tile(
        x: pa.Scalar,
        y: pa.Scalar,
        extent_x: pa.Scalar,
        extent_y: pa.Scalar,
        level: pa.Scalar,
    ) -> np.ndarray:
        rgba_region = slide.read_region_relative(
            (x.as_py(), y.as_py()),
            level.as_py(),
            (extent_x.as_py(), extent_y.as_py()),
        )
        rgb_region = Image.alpha_composite(
            Image.new("RGBA", rgba_region.size, (background, background, background)),
            rgba_region,
        ).convert("RGB")
        return np.asarray(rgb_region)

    return np.array(
        [
            get_tile(*args)
            for args in zip(
                tile_x,
                tile_y,
                tile_extent_x,
                tile_extent_y,
                level,
                strict=True,
            )
        ]
    )


def _read_tifffile_tiles(
    slide: TiffFile,
    tile_x: pa.IntegerArray,
    tile_y: pa.IntegerArray,
    tile_extent_x: pa.IntegerArray,
    tile_extent_y: pa.IntegerArray,
    level: pa.IntegerArray,
    background: int = 255,
) -> np.ndarray:
    """Read batch of tiles from an OME-TIFF file using tifffile.

    Args:
        slide: The OME-TIFF object to read tiles from.
        tile_x: Array of tile x-coordinates (in pixels) at level 0.
        tile_y: Array of tile y-coordinates (in pixels) at level 0.
        tile_extent_x: Array of tile widths in pixels.
        tile_extent_y: Array of tile heights in pixels.
        level: Array of OME-TIFF level indices for each tile.
        background: The RGB value (0-255) to use for transparent areas.
            Defaults to 255 (white).

    Returns:
        A NumPy array of RGB tiles read from the slide.
    """
    import tifffile
    import zarr
    from zarr.core.buffer import NDArrayLike

    def get_tile(
        z: zarr.Array,
        x: pa.Scalar,
        y: pa.Scalar,
        extent_x: pa.Scalar,
        extent_y: pa.Scalar,
    ) -> np.ndarray:
        x_py = x.as_py()
        y_py = y.as_py()
        extent_x_py: int = extent_x.as_py()
        extent_y_py: int = extent_y.as_py()

        arr = np.full((extent_y_py, extent_x_py, 3), background, dtype=np.uint8)
        tile_slice = z[
            y_py : y_py + extent_y_py,
            x_py : x_py + extent_x_py,
        ]
        assert isinstance(tile_slice, NDArrayLike)
        arr[: tile_slice.shape[0], : tile_slice.shape[1]] = tile_slice[..., :3]  # type: ignore[index]
        return arr

    # All the lists must have the same length
    tiles = np.empty(len(tile_x), dtype=object)

    for page, group in _pyarrow_group_indices(level).items():
        assert isinstance(page, int)
        page = slide.series[0].pages[page]
        assert isinstance(page, tifffile.TiffPage)

        z = zarr.open(page.aszarr(), mode="r")
        assert isinstance(z, zarr.Array)

        tiles[group] = [
            get_tile(z, *args)
            for args in zip(
                pc.take(tile_x, group),
                pc.take(tile_y, group),
                pc.take(tile_extent_x, group),
                pc.take(tile_extent_y, group),
                strict=True,
            )
        ]

    return tiles


def _pyarrow_group_indices(x: pa.Array) -> dict[Any, pa.Int32Array]:
    """Group indices of a PyArrow array by unique values.

    Args:
        x: A PyArrow array to group.

    Returns:
        A dictionary mapping unique values to PyArrow arrays of integer indices where those values occur.
    """
    unique_values = pc.unique(x)
    full_indices = pa.array(range(len(x)), type=pa.int32())

    groups = {}

    for value in unique_values:
        mask = pc.equal(x, value)
        groups[value.as_py()] = pc.filter(full_indices, mask)

    return groups



"""ASAP format annotation parser."""

import re
import xml.etree.ElementTree as ET
from collections.abc import Iterable
from pathlib import Path
from typing import TextIO

from shapely.geometry import Point, Polygon


class ASAPParser:
    """Parser for ASAP format annotation files.

    ASAP (Automated Slide Analysis Platform) uses XML format for storing annotations.
    This parser supports both polygon and point annotations.
    """

    def __init__(self, file_path: Path | str | TextIO):
        self.tree = ET.parse(file_path)
        self.root = self.tree.getroot()

    def _get_filtered_annotations(
        self, name: str, part_of_group: str
    ) -> Iterable[ET.Element]:
        """Get annotations that match the provided regex filters.

        Args:
            name: Regex pattern to match annotation names.
            part_of_group: Regex pattern to match annotation groups.

        Yields:
            XML annotation elements that match the filters.
        """
        name_regex = re.compile(name)
        part_of_group_regex = re.compile(part_of_group)

        for annotation in self.root.findall(".//Annotation"):
            if name_regex.match(
                annotation.attrib["Name"]
            ) and part_of_group_regex.match(annotation.attrib["PartOfGroup"]):
                yield annotation

    def _extract_coordinates(self, annotation: ET.Element) -> list[Point]:
        """Extract coordinates from an annotation element.

        Args:
            annotation: XML annotation element.

        Returns:
            List of (x, y) coordinate tuples.
        """
        return [
            Point(float(coordinate.attrib["X"]), float(coordinate.attrib["Y"]))
            for coordinate in annotation.findall(".//Coordinate")
        ]

    def get_polygons(
        self, name: str = ".*", part_of_group: str = ".*"
    ) -> Iterable[Polygon]:
        """Parse polygon annotations from ASAP XML file.

        Args:
            name: Regex pattern to match annotation names.
            part_of_group: Regex pattern to match annotation groups.

        Returns:
            An iterable of shapely Polygon objects.
        """
        for annotation in self._get_filtered_annotations(name, part_of_group):
            if annotation.attrib["Type"] in ["Polygon", "Spline"]:
                yield Polygon(self._extract_coordinates(annotation))

    def get_points(
        self, name: str = ".*", part_of_group: str = ".*"
    ) -> Iterable[Point]:
        """Parse point annotations from ASAP XML file.

        Args:
            name: Regex pattern to match annotation names.
            part_of_group: Regex pattern to match annotation groups.

        Returns:
            An iterable of shapely Point objects.
        """
        for annotation in self._get_filtered_annotations(name, part_of_group):
            if annotation.attrib["Type"] in ["Point", "Dot"]:
                yield from self._extract_coordinates(annotation)



from pathlib import Path

import numpy as np
import pandas as pd
from PIL import Image
from torch.utils.data import Dataset

from ratiopath.torch.data.openslide_tile_reader import OpenSlideTileReader


class SlideDataset(Dataset[pd.Series], OpenSlideTileReader):
    """Dataset for reading tiles from a single slide image.

    This dataset reads tiles from an OpenSlide image. The tiles are specified by a
    DataFrame with columns ["x", "y"]. The RGBA tiles are converted to RGB before
    being returned.

    Attributes:
        slide (Path): Path to the slide image.
        level (int | str): Level of the slide to read. If int, it is used as the level.
            If str, it is used as the column name in the tiles DataFrame.
        tile_extent_x (int | str): Width of the tile. If int, it is used as the width.
            If str, it is used as the column name in the tiles DataFrame.
        tile_extent_y (int | str): Height of the tile. If int, it is used as the height.
            If str, it is used as the column name in the tiles DataFrame.
        tiles (pd.DataFrame): DataFrame with columns ["x", "y"] specifying the tiles
            to be read.
    """

    def __init__(
        self,
        slide_tiles: pd.DataFrame,
        slide_path: str | Path,
        tile_extent: int | tuple[int, int] | str,
        level: int | str | None,
        slide_resolution: float | tuple[float, float] | None,
        background: None | tuple[int, int, int] = (255, 255, 255),
    ) -> None:
        """Initialize OpenSlideTilesDataset dataset.

        Args:
            slide_tiles: DataFrame with columns ["x", "y"] specifying the tiles to be read.
            slide_path: Path to the slide image.
            tile_extent: Width and height of the tile. If int, it is used as both the
                width and height. If str, it is used as the column name in the tiles
                DataFrame.
            level: Level of the slide to read. If int, it is used as the level. If str,
                it is used as the column name in the tiles DataFrame.
            slide_resolution: Mpp resolution.
            background: Background color for the tiles.
        """
        super().__init__(
            path=slide_path,
            level=level,
            resolution=slide_resolution,
            background=background,
        )

        self.slide_tiles = slide_tiles
        self.tile_extent = tile_extent

    def __len__(self) -> int:
        return len(self.slide_tiles)

    def __getitem__(self, idx: int) -> pd.Series:
        return self.slide_tiles.iloc[idx]

    def get_tile(self, tile_coords: tuple[int, int], tile: pd.Series) -> Image.Image:
        """Returns tile from the slide image at the specified coordinates in RGB format."""
        tile_extent = self._get_from_tile(tile, self.tile_extent)

        return super().get_openslide_tile(
            tile_coords=tile_coords,
            tile_extent=tuple(np.broadcast_to(tile_extent, 2)),  # type: ignore[call-overload]
            tile=tile,
        )



import numpy as np
import openslide
from openslide import PROPERTY_NAME_MPP_X, PROPERTY_NAME_MPP_Y
from PIL.Image import Image


class OpenSlide(openslide.OpenSlide):
    """A wrapper around the OpenSlide library to provide additional functionality."""

    def closest_level(self, mpp: float | tuple[float, float]) -> int:
        """Finds the closest slide level to match the desired MPP.

        This method compares the desired MPP (µm/px) with the MPP of the
        available levels in the slide and selects the level with the closest match.

        Args:
            mpp: The desired µm/px value.

        Returns:
            The index of the level with the closest µm/px resolution to the desired value.
        """
        scale_factor = np.mean(
            np.asarray(mpp)
            / np.array(
                [
                    float(self.properties[PROPERTY_NAME_MPP_X]),
                    float(self.properties[PROPERTY_NAME_MPP_Y]),
                ]
            )
        )

        return np.abs(np.asarray(self.level_downsamples) - scale_factor).argmin().item()

    def slide_resolution(self, level: int) -> tuple[float, float]:
        """Returns the resolution of the slide in µm/px at the given level.

        Args:
            level: The level of the slide to calculate the resolution.

        Returns:
            The [x, y] resolution of the slide in µm/px.
        """
        slide_mpp_x = float(self.properties[PROPERTY_NAME_MPP_X])
        slide_mpp_y = float(self.properties[PROPERTY_NAME_MPP_Y])

        return (
            slide_mpp_x * self.level_downsamples[level],
            slide_mpp_y * self.level_downsamples[level],
        )

    def read_region_relative(
        self, location: tuple[int, int], level: int, size: tuple[int, int]
    ) -> Image:
        """Reads a region from the slide with coordinates relative to the specified level.

        This method adjusts the coordinates based on the level's downsampling factor
        before reading the region from the slide.

        Args:
            location: The (x, y) coordinates at the specified level.
            level: The level of the slide to read from.
            size: The (width, height) of the region to read.

        Returns:
            The image of the requested region.
        """
        downsample = self.level_downsamples[level]
        location = (int(location[0] * downsample), int(location[1] * downsample))

        return super().read_region(location, level, size)



from ratiopath.parsers.asap_parser import ASAPParser
from ratiopath.parsers.darwin7_json_parser import Darwin7JSONParser
from ratiopath.parsers.geojson_parser import GeoJSONParser


__all__ = ["ASAPParser", "Darwin7JSONParser", "GeoJSONParser"]
